// Code generated by mockery v1.1.0. DO NOT EDIT.

package proxy

import mock "github.com/stretchr/testify/mock"

// MockPooler is an autogenerated mock type for the Pooler type
type MockPooler struct {
	mock.Mock
}

// Pool provides a mock function with given fields:
func (_m *MockPooler) Pool() Pool {
	ret := _m.Called()

	var r0 Pool
	if rf, ok := ret.Get(0).(func() Pool); ok {
		r0 = rf()
	} else {
		if ret.Get(0) != nil {
			r0 = ret.Get(0).(Pool)
		}
	}

	return r0
}
