// Code generated by mockery v1.1.0. DO NOT EDIT.

package volumes

import mock "github.com/stretchr/testify/mock"

// mockDebugLogger is an autogenerated mock type for the debugLogger type
type mockDebugLogger struct {
	mock.Mock
}

// Debugln provides a mock function with given fields: args
func (_m *mockDebugLogger) Debugln(args ...interface{}) {
	var _ca []interface{}
	_ca = append(_ca, args...)
	_m.Called(_ca...)
}
