// Code generated by mockery v1.1.0. DO NOT EDIT.

package process

import mock "github.com/stretchr/testify/mock"

// mockKiller is an autogenerated mock type for the killer type
type mockKiller struct {
	mock.Mock
}

// ForceKill provides a mock function with given fields:
func (_m *mockKiller) ForceKill() {
	_m.Called()
}

// Terminate provides a mock function with given fields:
func (_m *mockKiller) Terminate() {
	_m.Called()
}
