// Code generated by mockery v1.1.0. DO NOT EDIT.

package helpers

import mock "github.com/stretchr/testify/mock"

// mockLogOutputWriter is an autogenerated mock type for the logOutputWriter type
type mockLogOutputWriter struct {
	mock.Mock
}

// Write provides a mock function with given fields: _a0
func (_m *mockLogOutputWriter) Write(_a0 string) {
	_m.Called(_a0)
}
