package helpers

const (
	ANSI_BOLD_BLACK   = "\033[30;1m"
	ANSI_BOLD_RED     = "\033[31;1m"
	ANSI_BOLD_GREEN   = "\033[32;1m"
	ANSI_BOLD_YELLOW  = "\033[33;1m"
	ANSI_BOLD_BLUE    = "\033[34;1m"
	ANSI_BOLD_MAGENTA = "\033[35;1m"
	ANSI_BOLD_CYAN    = "\033[36;1m"
	ANSI_BOLD_WHITE   = "\033[37;1m"
	ANSI_YELLOW       = "\033[0;33m"
	ANSI_RESET        = "\033[0;m"
	ANSI_CLEAR        = "\033[0K"
)
