// Code generated by mockery v1.1.0. DO NOT EDIT.

package permission

import (
	context "context"

	mock "github.com/stretchr/testify/mock"
)

// MockSetter is an autogenerated mock type for the Setter type
type MockSetter struct {
	mock.Mock
}

// Set provides a mock function with given fields: ctx, volumeName
func (_m *MockSetter) Set(ctx context.Context, volumeName string) error {
	ret := _m.Called(ctx, volumeName)

	var r0 error
	if rf, ok := ret.Get(0).(func(context.Context, string) error); ok {
		r0 = rf(ctx, volumeName)
	} else {
		r0 = ret.Error(0)
	}

	return r0
}
