// Code generated by mockery v1.1.0. DO NOT EDIT.

package kubernetes

import (
	time "time"

	mock "github.com/stretchr/testify/mock"
)

// mockBackoffCalculator is an autogenerated mock type for the backoffCalculator type
type mockBackoffCalculator struct {
	mock.Mock
}

// ForAttempt provides a mock function with given fields: attempt
func (_m *mockBackoffCalculator) ForAttempt(attempt float64) time.Duration {
	ret := _m.Called(attempt)

	var r0 time.Duration
	if rf, ok := ret.Get(0).(func(float64) time.Duration); ok {
		r0 = rf(attempt)
	} else {
		r0 = ret.Get(0).(time.Duration)
	}

	return r0
}
