// Code generated by mockery v1.1.0. DO NOT EDIT.

package kubernetes

import (
	context "context"

	mock "github.com/stretchr/testify/mock"
)

// mockLogProcessor is an autogenerated mock type for the logProcessor type
type mockLogProcessor struct {
	mock.Mock
}

// Process provides a mock function with given fields: ctx
func (_m *mockLogProcessor) Process(ctx context.Context) <-chan string {
	ret := _m.Called(ctx)

	var r0 <-chan string
	if rf, ok := ret.Get(0).(func(context.Context) <-chan string); ok {
		r0 = rf(ctx)
	} else {
		if ret.Get(0) != nil {
			r0 = ret.Get(0).(<-chan string)
		}
	}

	return r0
}
