// Code generated by mockery v1.1.0. DO NOT EDIT.

package proxy

import (
	http "net/http"

	mock "github.com/stretchr/testify/mock"
)

// MockRequester is an autogenerated mock type for the Requester type
type MockRequester struct {
	mock.Mock
}

// ProxyRequest provides a mock function with given fields: w, r, requestedURI, port, settings
func (_m *MockRequester) ProxyRequest(w http.ResponseWriter, r *http.Request, requestedURI string, port string, settings *Settings) {
	_m.Called(w, r, requestedURI, port, settings)
}
