// Code generated by mockery v1.1.0. DO NOT EDIT.

package referees

import mock "github.com/stretchr/testify/mock"

// MockMetricsExecutor is an autogenerated mock type for the MetricsExecutor type
type MockMetricsExecutor struct {
	mock.Mock
}

// GetMetricsSelector provides a mock function with given fields:
func (_m *MockMetricsExecutor) GetMetricsSelector() string {
	ret := _m.Called()

	var r0 string
	if rf, ok := ret.Get(0).(func() string); ok {
		r0 = rf()
	} else {
		r0 = ret.Get(0).(string)
	}

	return r0
}
