// Code generated by mockery v1.1.0. DO NOT EDIT.

package common

import (
	mock "github.com/stretchr/testify/mock"
	cli "github.com/urfave/cli"
)

// MockCommander is an autogenerated mock type for the Commander type
type MockCommander struct {
	mock.Mock
}

// Execute provides a mock function with given fields: c
func (_m *MockCommander) Execute(c *cli.Context) {
	_m.Called(c)
}
