// Code generated by mockery v1.1.0. DO NOT EDIT.

package common

import mock "github.com/stretchr/testify/mock"

// MockFailuresCollector is an autogenerated mock type for the FailuresCollector type
type MockFailuresCollector struct {
	mock.Mock
}

// RecordFailure provides a mock function with given fields: reason, runnerDescription
func (_m *MockFailuresCollector) RecordFailure(reason JobFailureReason, runnerDescription string) {
	_m.Called(reason, runnerDescription)
}
