// Code generated by mockery v1.1.0. DO NOT EDIT.

package ca_chain

import mock "github.com/stretchr/testify/mock"

// mockFetcher is an autogenerated mock type for the fetcher type
type mockFetcher struct {
	mock.Mock
}

// Fetch provides a mock function with given fields: url
func (_m *mockFetcher) Fetch(url string) ([]byte, error) {
	ret := _m.Called(url)

	var r0 []byte
	if rf, ok := ret.Get(0).(func(string) []byte); ok {
		r0 = rf(url)
	} else {
		if ret.Get(0) != nil {
			r0 = ret.Get(0).([]byte)
		}
	}

	var r1 error
	if rf, ok := ret.Get(1).(func(string) error); ok {
		r1 = rf(url)
	} else {
		r1 = ret.Error(1)
	}

	return r0, r1
}
