// Code generated by mockery v1.1.0. DO NOT EDIT.

package common

import mock "github.com/stretchr/testify/mock"

// MockExecutorData is an autogenerated mock type for the ExecutorData type
type MockExecutorData struct {
	mock.Mock
}
