// Code generated by mockery v1.1.0. DO NOT EDIT.

package process

import mock "github.com/stretchr/testify/mock"

// MockKillWaiter is an autogenerated mock type for the KillWaiter type
type MockKillWaiter struct {
	mock.Mock
}

// KillAndWait provides a mock function with given fields: command, waitCh
func (_m *MockKillWaiter) KillAndWait(command Commander, waitCh chan error) error {
	ret := _m.Called(command, waitCh)

	var r0 error
	if rf, ok := ret.Get(0).(func(Commander, chan error) error); ok {
		r0 = rf(command, waitCh)
	} else {
		r0 = ret.Error(0)
	}

	return r0
}
