// Code generated by mockery v1.1.0. DO NOT EDIT.

package retry

import mock "github.com/stretchr/testify/mock"

// MockRetryable is an autogenerated mock type for the Retryable type
type MockRetryable struct {
	mock.Mock
}

// Run provides a mock function with given fields:
func (_m *MockRetryable) Run() error {
	ret := _m.Called()

	var r0 error
	if rf, ok := ret.Get(0).(func() error); ok {
		r0 = rf()
	} else {
		r0 = ret.Error(0)
	}

	return r0
}

// ShouldRetry provides a mock function with given fields: tries, err
func (_m *MockRetryable) ShouldRetry(tries int, err error) bool {
	ret := _m.Called(tries, err)

	var r0 bool
	if rf, ok := ret.Get(0).(func(int, error) bool); ok {
		r0 = rf(tries, err)
	} else {
		r0 = ret.Get(0).(bool)
	}

	return r0
}
